{% block imports %}
if 'callback' not in globals():
    from mage_ai.data_preparation.decorators import callback
{% endblock %}


{% block content %}
@callback('success')
def success_callback(**kwargs):
    pass


@callback('failure')
def failure_callback(**kwargs):
    pass
{% endblock %}
