"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[9307],{89307:function(e){e.exports=JSON.parse('{"base":"vs-dark","inherit":true,"rules":[{"foreground":"e6e1c4","background":"322323","token":"source"},{"foreground":"6b4e32","fontStyle":"italic","token":"comment"},{"foreground":"ef5d32","token":"keyword"},{"foreground":"ef5d32","token":"storage"},{"foreground":"efac32","token":"entity.name.function"},{"foreground":"efac32","token":"keyword.other.name-of-parameter.objc"},{"foreground":"efac32","fontStyle":"bold","token":"entity.name"},{"foreground":"6c99bb","token":"constant.numeric"},{"foreground":"7daf9c","token":"variable.language"},{"foreground":"7daf9c","token":"variable.other"},{"foreground":"6c99bb","token":"constant"},{"foreground":"efac32","token":"variable.other.constant"},{"foreground":"6c99bb","token":"constant.language"},{"foreground":"d9d762","token":"string"},{"foreground":"efac32","token":"support.function"},{"foreground":"efac32","token":"support.type"},{"foreground":"6c99bb","token":"support.constant"},{"foreground":"efcb43","token":"meta.tag"},{"foreground":"efcb43","token":"declaration.tag"},{"foreground":"efcb43","token":"entity.name.tag"},{"foreground":"efcb43","token":"entity.other.attribute-name"},{"foreground":"ffffff","background":"990000","token":"invalid"},{"foreground":"7daf9c","token":"constant.character.escaped"},{"foreground":"7daf9c","token":"constant.character.escape"},{"foreground":"7daf9c","token":"string source"},{"foreground":"7daf9c","token":"string source.ruby"},{"foreground":"e6e1dc","background":"144212","token":"markup.inserted"},{"foreground":"e6e1dc","background":"660000","token":"markup.deleted"},{"background":"2f33ab","token":"meta.diff.header"},{"background":"2f33ab","token":"meta.separator.diff"},{"background":"2f33ab","token":"meta.diff.index"},{"background":"2f33ab","token":"meta.diff.range"}],"colors":{"editor.foreground":"#E6E1C4","editor.background":"#372725","editor.selectionBackground":"#16120E","editor.lineHighlightBackground":"#1F1611","editorCursor.foreground":"#E6E1C4","editorWhitespace.foreground":"#42302D"}}')}}]);