"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5824],{15824:function(e){e.exports=JSON.parse('{"base":"vs","inherit":true,"rules":[{"background":"FFFFFF","token":""},{"foreground":"0066ff","fontStyle":"italic","token":"comment"},{"foreground":"bfbfbf","token":"deco.folding"},{"foreground":"0000ff","fontStyle":"bold","token":"keyword"},{"foreground":"0000ff","fontStyle":"bold","token":"storage"},{"foreground":"0000cd","token":"constant.numeric"},{"foreground":"c5060b","fontStyle":"bold","token":"constant"},{"foreground":"585cf6","fontStyle":"bold","token":"constant.language"},{"foreground":"318495","token":"variable.language"},{"foreground":"318495","token":"variable.other"},{"foreground":"036a07","token":"string"},{"foreground":"26b31a","token":"constant.character.escape"},{"foreground":"26b31a","token":"string meta.embedded"},{"foreground":"1a921c","token":"meta.preprocessor"},{"foreground":"0c450d","fontStyle":"bold","token":"keyword.control.import"},{"foreground":"0000a2","fontStyle":"bold","token":"entity.name.function"},{"foreground":"0000a2","fontStyle":"bold","token":"support.function.any-method"},{"fontStyle":"underline","token":"entity.name.type"},{"fontStyle":"italic","token":"entity.other.inherited-class"},{"fontStyle":"italic","token":"variable.parameter"},{"foreground":"70727e","token":"storage.type.method"},{"fontStyle":"italic","token":"meta.section entity.name.section"},{"fontStyle":"italic","token":"declaration.section entity.name.section"},{"foreground":"3c4c72","fontStyle":"bold","token":"support.function"},{"foreground":"6d79de","fontStyle":"bold","token":"support.class"},{"foreground":"6d79de","fontStyle":"bold","token":"support.type"},{"foreground":"06960e","fontStyle":"bold","token":"support.constant"},{"foreground":"21439c","fontStyle":"bold","token":"support.variable"},{"foreground":"687687","token":"keyword.operator.js"},{"foreground":"ffffff","background":"990000","token":"invalid"},{"background":"ffd0d0","token":"invalid.deprecated.trailing-whitespace"},{"background":"0000000d","token":"text source"},{"background":"0000000d","token":"string.unquoted"},{"background":"0000000d","token":"meta.embedded"},{"background":"0000000f","token":"text source string.unquoted"},{"background":"0000000f","token":"text source text source"},{"foreground":"68685b","token":"meta.tag.preprocessor.xml"},{"foreground":"888888","token":"meta.tag.metadata.doctype"},{"foreground":"888888","token":"meta.tag.metadata.doctype entity"},{"foreground":"888888","token":"meta.tag.metadata.doctype string"},{"foreground":"888888","token":"meta.tag.metadata.processing.xml"},{"foreground":"888888","token":"meta.tag.metadata.processing.xml entity"},{"foreground":"888888","token":"meta.tag.metadata.processing.xml string"},{"fontStyle":"italic","token":"meta.tag.metadata.doctype string.quoted"},{"foreground":"1c02ff","token":"meta.tag"},{"foreground":"1c02ff","token":"declaration.tag"},{"fontStyle":"bold","token":"entity.name.tag"},{"fontStyle":"italic","token":"entity.other.attribute-name"},{"foreground":"0c07ff","fontStyle":"bold","token":"markup.heading"},{"foreground":"000000","fontStyle":"italic","token":"markup.quote"},{"foreground":"b90690","token":"markup.list"}],"colors":{"editor.foreground":"#000000","editor.background":"#FFFFFF","editor.selectionBackground":"#4D97FF54","editor.lineHighlightBackground":"#00000012","editorCursor.foreground":"#000000","editorWhitespace.foreground":"#BFBFBF"}}')}}]);