{% block imports %}
{% endblock %}
if 'test' not in globals():
    from mage_ai.data_preparation.decorators import test


{% block content %}
{% endblock %}


@test
def test_output(output, *args) -> None:
    """
    Template code for testing the output of the block.
    """
    assert output is not None, 'The output is undefined'
