"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[4450],{4450:function(e){e.exports=JSON.parse('{"base":"vs-dark","inherit":true,"rules":[{"background":"000000","token":""},{"foreground":"ffffff","background":"434242","token":"text"},{"foreground":"ffffff","background":"000000","token":"source"},{"foreground":"9933cc","token":"comment"},{"foreground":"3387cc","token":"constant"},{"foreground":"cc7833","token":"keyword"},{"foreground":"d0d0ff","token":"meta.preprocessor.c"},{"fontStyle":"italic","token":"variable.parameter"},{"foreground":"ffffff","background":"9b9b9b","token":"source comment.block"},{"foreground":"66cc33","token":"string"},{"foreground":"aaaaaa","token":"string constant.character.escape"},{"foreground":"000000","background":"cccc33","token":"string.interpolated"},{"foreground":"cccc33","token":"string.regexp"},{"foreground":"cccc33","token":"string.literal"},{"foreground":"555555","token":"string.interpolated constant.character.escape"},{"fontStyle":"underline","token":"entity.name.type"},{"fontStyle":"italic underline","token":"entity.other.inherited-class"},{"fontStyle":"underline","token":"entity.name.tag"},{"foreground":"c83730","token":"support.function"}],"colors":{"editor.foreground":"#FFFFFF","editor.background":"#000000","editor.selectionBackground":"#73597EE0","editor.lineHighlightBackground":"#333300","editorCursor.foreground":"#FFFFFF","editorWhitespace.foreground":"#404040"}}')}}]);