(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5405],{63878:function(n,u,t){"use strict";t.r(u);var e=t(82684),i=t(34376),s=t(82531);u.default=function(){var n=(0,i.useRouter)(),u=n.asPath,t=u.split("?")[0],a=s.ZP.statuses.list().data,o=(0,e.useMemo)((function(){var n;return null===a||void 0===a||null===(n=a.statuses)||void 0===n?void 0:n[0]}),[a]);(0,e.useEffect)((function(){if(o){var e=null===o||void 0===o?void 0:o.is_instance_manager,i=u;"/"===t&&(i=e?"/manage":"/pipelines"),n.replace(i)}}),[t,u,o,n])}},48312:function(n,u,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/",function(){return t(63878)}])}},function(n){n.O(0,[9774,2888,179],(function(){return u=48312,n(n.s=u);var u}));var u=n.O();_N_E=u}]);