"""Add completed_at.

Revision ID: 053ee2c10d85
Revises: 52ab80005742
Create Date: 2022-08-31 01:22:18.258615

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '053ee2c10d85'
down_revision = '52ab80005742'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('block_run', sa.Column('completed_at', sa.DateTime(timezone=True), nullable=True))
    op.add_column('pipeline_run', sa.Column('completed_at', sa.DateTime(timezone=True), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipeline_run', 'completed_at')
    op.drop_column('block_run', 'completed_at')
    # ### end Alembic commands ###
