{% extends "transformers/default.jinja" %}
{% block imports %}
from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import build_transformer_action
{{ super() -}}
{% endblock %}


{% block content %}
@transformer
def {{ title }}(df: DataFrame, *args, **kwargs) -> DataFrame:
    """
    Transformer Action: {{ message }}
    """
    action = {{ payload }}
    return BaseAction(action).execute(df)
{% endblock %}
